// mml2fssql

#include <stdio.h>
#include <malloc.h>
#include <memory.h>
#include <ctype.h>
#include <string.h>

#define CH_MAX         4
#define EVENT_MAX   4096
#define LABEL_MAX   1024
#define LENTBL_SIZE   15
#define LENTBL_MAX   256
#define TONE_RANGE    14
#define ENV_MAX       15
#define ENV_SIZE     128
#define RP_NAME_SIZE  64
#define RP_MAX        32

#define BASE_TICK 192

#define TYPE_UNKNOWN    0
#define TYPE_NOTE       1
#define TYPE_REST       2
#define TYPE_EXTEND     3
#define TYPE_REG1       4
#define TYPE_REG2       5
#define TYPE_TONE_BASE  6
#define TYPE_LEN_TABLE  7
#define TYPE_ENV_BASE   8
#define TYPE_TEMPO      9
#define TYPE_PERCUSS   10
#define TYPE_VOLUME    11
#define TYPE_REPEAT    12
#define TYPE_REPEAT_W  13
#define TYPE_JMP       14
#define TYPE_CALL      15
#define TYPE_RET       16
#define TYPE_LOOP_CNT  17
#define TYPE_END       18
#define TYPE_LABEL     32
#define TYPE_SKIP      33
#define TYPE_SKIP_W    34

#define DEF_EP 0x00004550
#define DEF_RP 0x00005250

#define DEF_TEMPO 120

//           value        len        quant
// NOTE                      quantl
// REST      0                   quantl
// EXTEND    0                   quantl
// REG1      ݒl       0          0
// REG2      ݒl       0          0
// TONE_BASE ()   () 0
// LEN_TABLE LTableԍ   0          0
// ENV_BASE  envԍ      0          0
// TEMPO     e|l     0          0
// PERCUSS   ŉԍ     0          0
// VOLUME    ύX       0          0
// REPEAT    JԂ xԍ short/long
// REPEAT_W  JԂ xԍ short/long
// JMP       QƔԍ     xԍ short/long
// CALL      QƔԍ     xԍ short/long
// RET       0            0          0
// LOOP_CNT  0            0          0
// END       0            0          0
// LABEL     QƔԍ     xԍ 0
// SKIP      0            xԍ 0
//
// xԍ gbN[v:0 ̑1ȏ

typedef struct {
	unsigned short next;
	unsigned char  type;
	unsigned char  quant;
	unsigned short value;
	unsigned short len;
} event_t;

typedef struct {
	unsigned short head;
	unsigned short last;
	unsigned short last_note;
	int base_tone;
	int def_len;
	int quant;
	int label;
	int save_base_tone;
	int save_base_tone_w;
	int save_def_len;
	int save_def_len_w;
	int save_quant;
	int save_quant_w;
	int exist_skip;
	int exist_skip_w;
	int assigned_label;
	int assigned_label_w;
	int nest;
	int nest_w;
	int ref_ev;
} seq_t;

typedef struct {
	unsigned short len[LENTBL_SIZE];
	unsigned char  quant[LENTBL_SIZE];
	int count;
	int base;
	int share_cnt;
} lentbl_t;

typedef struct {
	int value;
} env_t;

typedef struct {
	char name[RP_NAME_SIZE];
	unsigned short data;
	int quant;
} rhythm_t;

typedef struct {
	event_t  events[EVENT_MAX];
	lentbl_t lentbl[LENTBL_MAX];
	env_t    env[ENV_MAX];
	rhythm_t rhythm[RP_MAX];
	seq_t    ch[CH_MAX];
	int evcnt;
	int ltcnt;
	int rpcnt;
	int tempo_1st;
	int tempo_max;
} score_t;

typedef struct {
	int pos;
	int count;
} call_t;

typedef struct {
	int pos;
	int low;
	int high;
	int ltbl;
} flow_t;

typedef struct {
	flow_t near;
	flow_t far;
	int link1;
	int link2;
} opttbl_t;

int parse(score_t *score, FILE *fp);
int insert_loop_event(score_t *score);
int insert_label_event(score_t *score);
int insert_tone_event(score_t *score);
int insert_lentbl_event(score_t *score);
int encode_envtable(score_t *score);
int optimize_event(score_t *score);
int assign_label_offset(score_t *score);
int assign_envtable_base(score_t *score);
int write_cmd(score_t *score, FILE *fp, const char *prefix);
int write_data(score_t *score, FILE *fp, const char *prefix);
void dump(score_t *score);

int main(int argc, char *argv[]) {
	FILE *ifp = NULL;
	FILE *cfp = NULL;
	FILE *dfp = NULL;
	const char *mml_name;
	const char *cmd_name = "out_cmd.asm";
	const char *dat_name = "out_data.asm";
	const char *prefix = "Out";
	score_t *score;
	int i;

	if (argc < 2) {
		printf("Usage: %s <mml file> [<cmd source file>] [<data source file>] [<prefix>]\n", argv[0]);
	}

	score = (score_t *)malloc(sizeof(score_t));
	if (score == NULL) goto fail;
	memset(score, 0, sizeof(score_t));

	mml_name = argv[1];
	if (argc >= 3) {
		cmd_name = argv[2];
	}
	if (argc >= 4) {
		dat_name = argv[3];
	}
	if (argc >= 5) {
		prefix = argv[4];
	}

	ifp = fopen(mml_name, "r");
	if (ifp == NULL) goto fail;

	cfp = fopen(cmd_name, "w");
	if (cfp == NULL) goto fail;

	dfp = fopen(dat_name, "w");
	if (dfp == NULL) goto fail;

	parse(score, ifp);
	insert_loop_event(score);
	insert_label_event(score);
	insert_tone_event(score);
	insert_lentbl_event(score);
	optimize_event(score);
	assign_label_offset(score);
	write_cmd(score, cfp, prefix);
	write_data(score, dfp, prefix);
//	dump(score);

fail:
	if (ifp) fclose(ifp);
	if (cfp) fclose(cfp);
	if (dfp) fclose(dfp);
	free(score);

	return 0;
}

void append_event(score_t *score, int i, int type, int value, int len, int quant) {
	event_t *ev;

	score->evcnt++;
	ev = score->events + score->evcnt;

	if (score->ch[i].head == 0) {
		score->ch[i].head = score->evcnt;
		score->ch[i].last = score->evcnt;
	} else {
		score->events[score->ch[i].last].next = score->evcnt;
		score->ch[i].last = score->evcnt;
	}

	ev->next  = 0;
	ev->type  = type;
	ev->value = value;
	ev->len   = len;
	ev->quant = quant;
}

void insert_event(score_t *score, int i, int prev, int type, int value, int len, int quant) {
	event_t *ev;

	score->evcnt++;
	ev = score->events + score->evcnt;

	if (score->ch[i].last == prev) {
		score->ch[i].last = score->evcnt;
	}
	if (prev == 0) {
		ev->next = score->ch[i].head;
		score->ch[i].head = score->evcnt;
	} else {
		ev->next = score->events[prev].next;
		score->events[prev].next = score->evcnt;
	}

	ev->type  = type;
	ev->value = value;
	ev->len   = len;
	ev->quant = quant;
}

char *parse_int(char *p, int *result, int num) {
	int value = 0;

	if (*p == '$') {
		p++;
		while (isxdigit(*p)) {
			if (isdigit(*p)) {
				value = value * 16 + (*p - '0');
			} else if (isupper(*p)) {
				value = value * 16 + (*p - 'A' + 10);
			} else {
				value = value * 16 + (*p - 'a' + 10);
			}
			p++;
		}
	} else {
		while (isdigit(*p)) {
			value = value * 10 + (*p - '0');
			p++;
		}
	}

	*result = value;
	return p;
}

char *parse_signed_int(char *p, int *result, int num) {
	int value = 0;
	int minus = 0;

	if (*p == '+') p++;
	if (*p == '-') {
		minus = 1;
		p++;
	}

	p = parse_int(p, &value, num);
	if (minus) value = -value;

	*result = value;
	return p;
}

char *parse_len(char *p, int *result, int def_len, int num) {
	char err_buf[64];
	char *q;
	int len, len2, len3;
	int i;

	q = p;
	len3 = 0;
	for (;;) {
		if (isdigit(*p) || (*p == '$')) {
			p = parse_int(p, &len, num);
			if (BASE_TICK % len) goto err_len;
			len = BASE_TICK / len;
		} else {
			len = BASE_TICK / def_len;
		}
		len2 = len;
		while (*p == '.') {
			if (len2 % 2) goto err_len;
			len2 /= 2;
			len += len2;
			p++;
		}
		len3 += len;

		if (*p != '^') break;
		p++;
	}

	*result = len3;
	return p;

err_len:
	p = q;
	q = err_buf;
	for (;;) {
		if (*p == '$') {
			*(q++) = *(p++);
			while (isxdigit(*p)) *(q++) = *(p++);
		} else {
			while (isdigit(*p)) *(q++) = *(p++);
		}
		while (*p == '.')  *(q++) = *(p++);
		if (*p != '^') break;
		p++;
	}
	*q = '\0';
	printf("%d: invalid tone length (%s)\n", num, err_buf);
	return p;
}

int parse_define(score_t *score, int def, char *p, int num) {
	char buf[4];
	char *q;
	int index;
	int value, tc, dt;
	int n;

	if (def == DEF_EP) {
		while (isspace(*p)) p++;
		p = parse_int(p, &index, num);
		if ((index >= 1) && (index <= ENV_MAX)) {
			while (isspace(*p)) p++;
			if (*p == ',') p++;
			while (isspace(*p)) p++;
			p = parse_int(p, &value, num);
			score->env[index-1].value = value;
		} else {
			printf("%d: invalid index of EP.\n", num);
		}
	} else if (def == DEF_RP) {
		if (score->rpcnt < RP_MAX) {
			index = score->rpcnt++;
			score->rhythm[index].data = 0;

			while (isspace(*p)) p++;
			if (*p == ',') p++;
			while (isspace(*p)) p++;
			q = score->rhythm[index].name;
			while (isalpha(*p) || (*p == '_')) {
				*(q++) = toupper(*p++);
			}
			*q = '\0';

			while (isspace(*p)) p++;
			if (*p == ',') p++;
			while (isspace(*p)) p++;
			p = parse_int(p, &value, num);
			score->rhythm[index].data |= ((value & 3) << 2);

			while (isspace(*p)) p++;
			if (*p == ',') p++;
			while (isspace(*p)) p++;
			p = parse_int(p, &value, num);
			switch (value) {
			case 1:
				value = 1;
				break;
			case 6:
				value = 2;
				break;
			case 7:
				value = 3;
				break;
			default:
				value = 0;
			}
			score->rhythm[index].data |= value;

			while (isspace(*p)) p++;
			if (*p == ',') p++;
			while (isspace(*p)) p++;
			p = parse_int(p, &value, num);
			score->rhythm[index].data |= ((value & 15) << 4);

			while (isspace(*p)) p++;
			if (*p == ',') p++;
			while (isspace(*p)) p++;
			if (isdigit(*p)) {
				p = parse_int(p, &value, num);
				score->rhythm[index].quant = value;
			} else {
				score->rhythm[index].quant = 0;
			}
		} else {
			printf("%d: Rhythm define is over.\n", num);
		}
	} else {
		// `
		q = buf;
		if (def & 0x00ff0000) *(q++) = (def>>16) & 0xff;
		if (def & 0x0000ff00) *(q++) = (def>> 8) & 0xff;
		if (def & 0x000000ff) *(q++) = (def    ) & 0xff;
		*q = '\0';
		printf("%d: syntax error. (#%s)\n", num, q);
	}

	return 0;
}

int parse_score(score_t *score, int ch, char *p, int num) {
	static int tone_inx[] = { 9, 11, 0, 2, 4, 5, 7 };
	char name[RP_NAME_SIZE];
	event_t *ev;
	char *q;
	int value;
	int len, len2;
	int nest;
	int i;

	while (*p) {
		if (isspace(*p)) {
			p++;
		} else if (*p == ';') {
			// Rg
			while (*p) p++;
		} else if (*p == '@') {
			p++;
			if (*p == 'x') {
				// WX^1w
				p++;
				if (isdigit(*p) || (*p == '$')) {
					p = parse_int(p, &value, num);
					if ((value >= 0) && (value < 256)) {
						for (i = 0; i < CH_MAX; i++) {
							if (ch & (1 << i)) {
								append_event(score, i, TYPE_REG1, value, 0, 0);
							}
						}
					} else {
						printf("%d: invalid value in reg1.\n", num);
					}
				} else {
					printf("%d: no value in reg1.\n", num);
				}
			} else if (*p == 'y') {
				// WX^2w
				p++;
				if (isdigit(*p) || (*p == '$')) {
					p = parse_int(p, &value, num);
					if ((value >= 0) && (value < 256)) {
						for (i = 0; i < CH_MAX; i++) {
							if (ch & (1 << i)) {
								append_event(score, i, TYPE_REG2, value, 0, 0);
							}
						}
					} else {
						printf("%d: invalid value in reg2.\n", num);
					}
				} else {
					printf("%d: no value in reg2.\n", num);
				}
			} else if ((*p == 'e') && (*(p+1) == 'v')) {
				// Gx[vw
				p += 2;
				if (isdigit(*p) || (*p == '$')) {
					p = parse_int(p, &value, num);
					if ((value >= 0) && (value <= ENV_MAX)) {
						for (i = 0; i < CH_MAX; i++) {
							if (ch & (1 << i)) {
								append_event(score, i, TYPE_ENV_BASE, value, 0, 0);
							}
						}
					} else {
						printf("%d: invalid value in volume envelope.\n", num);
					}
				} else {
					printf("%d: no value in volume envelope.\n", num);
				}
			} else if ((*p == 'l') && (*(p+1) == 'c')) {
				// [vJE^
				for (i = 0; i < CH_MAX; i++) {
					if (ch & (1 << i)) {
						append_event(score, i, TYPE_LOOP_CNT, 0, 0, 0);
					}
				}
				p += 2;
			} else if (*p == '[') {
				// L惊s[g(Jn)
				p++;
				for (i = 0; i < CH_MAX; i++) {
					if (ch & (1 << i)) {
						if (score->ch[i].nest_w < 1) {
							len  = score->ch[i].label;
							score->ch[i].assigned_label_w = len;
							score->ch[i].exist_skip_w = 0;
							append_event(score, i, TYPE_LABEL, 0, len, 0);
							score->ch[i].label++;
							score->ch[i].nest_w++;
						} else {
							printf("%d: repeat nest over.\n", num);
						}
					}
				}
			} else if (*p == '|') {
				// L惊s[g(XLbv)
				p++;
				for (i = 0; i < CH_MAX; i++) {
					if (ch & (1 << i)) {
						len  = score->ch[i].assigned_label_w;
						append_event(score, i, TYPE_SKIP_W, 0, len, 0);
						score->ch[i].save_base_tone_w = score->ch[i].base_tone;
						score->ch[i].save_def_len_w   = score->ch[i].def_len;
						score->ch[i].save_quant_w     = score->ch[i].quant;
						score->ch[i].exist_skip_w     = 1;
					}
				}
			} else if (*p == ']') {
				// L惊s[g(I)
				p++;
				if (isdigit(*p) || (*p == '$')) {
					p = parse_int(p, &value, num);
					if ((value > 0) && (value <= 127)) {
						for (i = 0; i < CH_MAX; i++) {
							if (ch & (1 << i)) {
								score->ch[i].nest_w--;
								len  = score->ch[i].assigned_label_w;
								append_event(score, i, TYPE_REPEAT_W, value, len, 0);
								if (score->ch[i].exist_skip_w != 0) {
									score->ch[i].base_tone = score->ch[i].save_base_tone_w;
									score->ch[i].def_len   = score->ch[i].save_def_len_w;
									score->ch[i].quant     = score->ch[i].save_quant_w;
								}
							}
						}
					} else {
						printf("%d: invalid value in repeat.\n", num);
					}
				} else {
					printf("%d: no value in repeat.\n", num);
				}
			} else {
				// `
				printf("%d: syntax error. (@%c)\n", num, *p);
				p++;
			}
		} else if (strchr("cdefgab", *p)) {
			// 
			value = tone_inx[*p - 'a'];
			p++;
			if (*p == '+') {
				value++;
				p++;
			} else if (*p == '-') {
				value--;
				p++;
			}

			q = p;
			for (i = 0; i < CH_MAX; i++) {
				if (ch & (1 << i)) {
					p = parse_len(q, &len, score->ch[i].def_len, num);
					if (len > BASE_TICK) {
						len2 = len;
						while (len2 > BASE_TICK) len2 -= BASE_TICK;
						append_event(score, i, TYPE_NOTE, value + score->ch[i].base_tone, len2, 8);
						len2 += BASE_TICK;
						while (len2 < len) {
							append_event(score, i, TYPE_EXTEND, 0, BASE_TICK, 8);
							len2 += BASE_TICK;
						}
						append_event(score, i, TYPE_EXTEND, 0, BASE_TICK, score->ch[i].quant);
					} else {
						append_event(score, i, TYPE_NOTE, value + score->ch[i].base_tone, len, score->ch[i].quant);
					}
					score->ch[i].last_note = score->ch[i].last;
				}
			}
			if (score->tempo_1st == 0) score->tempo_1st = DEF_TEMPO;
			if (score->tempo_max == 0) score->tempo_max = DEF_TEMPO;
		} else if (*p == 'r') {
			// x
			p++;

			q = p;
			for (i = 0; i < CH_MAX; i++) {
				if (ch & (1 << i)) {
					p = parse_len(q, &len, score->ch[i].def_len, num);
					while (len > BASE_TICK) {
						append_event(score, i, TYPE_REST, 0, BASE_TICK, 8);
						len -= BASE_TICK;
					}
					append_event(score, i, TYPE_REST, 0, len, score->ch[i].quant);
					score->ch[i].last_note = score->ch[i].last;
				}
			}
			if (score->tempo_1st == 0) score->tempo_1st = DEF_TEMPO;
			if (score->tempo_max == 0) score->tempo_max = DEF_TEMPO;
		} else if (*p == '!') {
			// Y
			p++;

			q = name;
			while (isalpha(*p) || (*p == '_')) {
				*(q++) = toupper(*p++);
			}
			*q = '\0';

			value = -1;
			for (i = 0; i < score->rpcnt; i++) {
				if (strcmp(name, score->rhythm[i].name) == 0) {
					value = i;
					break;
				}
			}

			if (value >= 0) {
				q = p;
				if (ch & (1 << 3)) {
					p = parse_len(q, &len, score->ch[3].def_len, num);
					if (score->rhythm[value].quant > 0) {
						append_event(score, 3, TYPE_NOTE, value+32+2*12, len, score->rhythm[value].quant);
					} else {
						append_event(score, 3, TYPE_NOTE, value+32+2*12, len, score->ch[3].quant);
					}
					score->ch[3].last_note = score->ch[3].last;
				}
			} else {
				printf("%d: undefined rhythm name. (%s)\n", num, name);
			}
			if (score->tempo_1st == 0) score->tempo_1st = DEF_TEMPO;
			if (score->tempo_max == 0) score->tempo_max = DEF_TEMPO;
		} else if (*p == '&') {
			// ^C
			p++;

			q = p;
			for (i = 0; i < CH_MAX; i++) {
				if (ch & (1 << i)) {
					p = parse_len(q, &len, score->ch[i].def_len, num);
					if (score->ch[i].last_note > 0) {
						ev = &score->events[score->ch[i].last_note];
						ev->quant = 8;
						if (len > BASE_TICK) {
							len2 = len;
							while (len2 > BASE_TICK) len2 -= BASE_TICK;
							append_event(score, i, TYPE_EXTEND, 0, len2, 8);
							len2 += BASE_TICK;
							while (len2 < len) {
								append_event(score, i, TYPE_EXTEND, 0, BASE_TICK, 8);
								len2 += BASE_TICK;
							}
							append_event(score, i, TYPE_EXTEND, 0, BASE_TICK, score->ch[i].quant);
						} else {
							append_event(score, i, TYPE_EXTEND, 0, len, score->ch[i].quant);
						}
						score->ch[i].last_note = score->ch[i].last;
						//append_event(score, i, TYPE_TIE, 0, 0, score->ch[i].quant);
						//ev = &score->events[score->ch[i].last_note];
						//append_event(score, i, ev->type, ev->value, len, score->ch[i].quant);
					}
				}
			}
		} else if (*p == '>') {
			// IN^[u㏸
			for (i = 0; i < CH_MAX; i++) {
				if (ch & (1 << i)) {
					score->ch[i].base_tone += 12;
				}
			}
			p++;
		} else if (*p == '<') {
			// IN^[u~
			for (i = 0; i < CH_MAX; i++) {
				if (ch & (1 << i)) {
					score->ch[i].base_tone -= 12;
				}
			}
			p++;
		} else if (*p == 'o') {
			// IN^[uw
			p++;
			if (isdigit(*p) || (*p == '$')) {
				p = parse_int(p, &value, num);
				if ((value >= 2) && (value < 8)) {
					for (i = 0; i < CH_MAX; i++) {
						if (ch & (1 << i)) {
							score->ch[i].base_tone = value * 12;
						}
					}
				} else {
					printf("%d: invalid value in octave.\n", num);
				}
			} else {
				printf("%d: no value in octave.\n", num);
			}
		} else if (*p == 't') {
			// e|w
			p++;
			if (isdigit(*p) || (*p == '$')) {
				p = parse_int(p, &value, num);
				len = (score->tempo_1st == 0) ? 1 : 0;
				for (i = 0; i < CH_MAX; i++) {
					if (ch & (1 << i)) {
						append_event(score, i, TYPE_TEMPO, value, 0, len);
					}
				}
				if (score->tempo_1st == 0) score->tempo_1st = value;
				if (value > score->tempo_max) score->tempo_max = value;
			} else {
				printf("%d: no value in tempo.\n", num);
			}
		} else if (*p == 'l') {
			// ȗ̉w
			p++;
			if (isdigit(*p) || (*p == '$')) {
				p = parse_int(p, &value, num);
				if ((BASE_TICK % value) == 0) {
					for (i = 0; i < CH_MAX; i++) {
						if (ch & (1 << i)) {
							score->ch[i].def_len = value;
						}
					}
				} else {
					printf("%d: invalid value in default length.\n", num);
				}
			} else {
				printf("%d: no value in default length.\n", num);
			}
		} else if (*p == 'q') {
			// NH^CYw
			p++;
			if (isdigit(*p) || (*p == '$')) {
				p = parse_int(p, &value, num);
				if ((value > 0) && (value <= 8)) {
					for (i = 0; i < CH_MAX; i++) {
						if (ch & (1 << i)) {
							score->ch[i].quant = value;
						}
					}
				} else {
					printf("%d: invalid value in quantize.\n", num);
				}
			} else {
				printf("%d: no value in quantize.\n", num);
			}
		} else if (*p == 'm') {
			// R[
			p++;
			if (isdigit(*p) || (*p == '$')) {
				p = parse_int(p, &value, num);
				if ((value > 0) && (value <= 127)) {
					for (i = 0; i < CH_MAX; i++) {
						if (ch & (1 << i)) {
							append_event(score, i, TYPE_CALL, value, 0, 0);
						}
					}
				} else {
					printf("%d: invalid value in mimic.\n", num);
				}
			} else {
				printf("%d: no value in mimic.\n", num);
			}
		} else if (*p == ':') {
			// gbN[v
			p++;
			for (i = 0; i < CH_MAX; i++) {
				if (ch & (1 << i)) {
					append_event(score, i, TYPE_LABEL, 0, 0, 0);
				}
			}
		} else if (*p == '[') {
			// s[g(Jn)
			p++;
			for (i = 0; i < CH_MAX; i++) {
				if (ch & (1 << i)) {
					if (score->ch[i].nest < 1) {
						len  = score->ch[i].label;
						score->ch[i].assigned_label = len;
						score->ch[i].exist_skip = 0;
						append_event(score, i, TYPE_LABEL, 0, len, 0);
						score->ch[i].label++;
						score->ch[i].nest++;
					} else {
						printf("%d: repeat nest over.\n", num);
					}
				}
			}
		} else if (*p == '|') {
			// s[g(XLbv)
			p++;
			for (i = 0; i < CH_MAX; i++) {
				if (ch & (1 << i)) {
					len  = score->ch[i].assigned_label;
					append_event(score, i, TYPE_SKIP, 0, len, 0);
					score->ch[i].save_base_tone = score->ch[i].base_tone;
					score->ch[i].save_def_len   = score->ch[i].def_len;
					score->ch[i].save_quant     = score->ch[i].quant;
					score->ch[i].exist_skip     = 1;
				}
			}
		} else if (*p == ']') {
			// s[g(I)
			p++;
			if (isdigit(*p) || (*p == '$')) {
				p = parse_int(p, &value, num);
				if ((value > 0) && (value <= 127)) {
					for (i = 0; i < CH_MAX; i++) {
						if (ch & (1 << i)) {
							score->ch[i].nest--;
							len  = score->ch[i].assigned_label;
							append_event(score, i, TYPE_REPEAT, value, len, 0);
							if (score->ch[i].exist_skip != 0) {
								score->ch[i].base_tone = score->ch[i].save_base_tone;
								score->ch[i].def_len   = score->ch[i].save_def_len;
								score->ch[i].quant     = score->ch[i].save_quant;
							}
						}
					}
				} else {
					printf("%d: invalid value in repeat.\n", num);
				}
			} else {
				printf("%d: no value in repeat.\n", num);
			}
		} else if (*p == '(') {
			// QƔ͈(Jn)
			p++;
			for (i = 0; i < CH_MAX; i++) {
				if (ch & (1 << i)) {
					len  = score->ch[i].label;
					append_event(score, i, TYPE_LABEL, 0, len, 0);
					score->ch[i].ref_ev = score->evcnt;
					score->ch[i].label++;
				}
			}
		} else if (*p == ')') {
			// QƔ͈(I)
			p++;
			if (isdigit(*p) || (*p == '$')) {
				p = parse_int(p, &value, num);
				if ((value >= 1) && (value <= 127)) {
					for (i = 0; i < CH_MAX; i++) {
						if (ch & (1 << i)) {
							if (score->ch[i].ref_ev >= 0) {
								len = score->events[score->ch[i].ref_ev].len;
								append_event(score, i, TYPE_RET, 0, len, 0);
								score->events[score->ch[i].ref_ev].value = value;
								score->ch[i].ref_ev = -1;
							}
						}
					}
				} else {
					printf("%d: invalid value in reference.\n", num);
				}
			} else {
				printf("%d: no value in repeat.\n", num);
			}
		} else if (*p == '/') {
			// gbNI[
			p++;
			for (i = 0; i < CH_MAX; i++) {
				if (ch & (1 << i)) {
					append_event(score, i, TYPE_END, 0, 0, 0);
				}
			}
		} else {
			// `
			printf("%d: syntax error. (%c)\n", num, *p);
			p++;
		}
	}

	return 0;
}

int parse(score_t *score, FILE *fp) {
	char line[1024];
	char *p;
	int ch;
	int def;
	int num;
	int i;

	score->tempo_1st = 0;
	score->tempo_max = 0;
	for (i = 0; i < CH_MAX; i++) {
		score->ch[i].base_tone = 4*12;
		score->ch[i].def_len   = 4;
		score->ch[i].quant     = 7;
		score->ch[i].label     = 1;
		score->ch[i].nest      = 0;
		score->ch[i].ref_ev    = -1;
	}

	num = 0;
	while (!feof(fp)) {
		p = fgets(line, sizeof(line), fp);
		if (p == NULL) break;
		num++;

		ch  = 0;
		def = 0;
		while (*p) {
			if (isspace(*p)) {
				if (ch) break;
				p++;
			} else if (*p == ';') {
				// Rg
				while (*p) p++;
			} else if (strchr("ABCD", *p)) {
				// `lw
				ch |= 1 << (*p - 'A');
				p++;
			} else if (*p == '#') {
				// f[^`
				p++;
				i = 0;
				while ((i < 3) && isalnum(*p)) {
					def = (def << 8) | *p;
					p++;
					i++;
				}
				if (isspace(*p)) {
					break;
				} else {
					printf("%d: unknown define name.)\n", num);
				}
			} else {
				printf("%d: unknown channel. (%c)\n", num, *p);
				p++;
			}
		}

		if (def != 0) {
			parse_define(score, def, p, num);
		} else {
			parse_score(score, ch, p, num);
		}
	}

	return 0;
}

int insert_loop_event(score_t *score) {
	int head[LABEL_MAX];
	int scnd[LABEL_MAX];
	int skip[LABEL_MAX];
	int tail[LABEL_MAX];
	event_t *ev;
	event_t *end_ev;
	int pos, prev;
	int pos_head, pos_tail;
	int track_loop;
	int i, k;

	for (i = 0; i < CH_MAX; i++) {
		memset(head, 0, sizeof(head));
		memset(scnd, 0, sizeof(scnd));
		memset(skip, 0, sizeof(skip));
		memset(tail, 0, sizeof(tail));

		track_loop = 0;
		ev     = NULL;
		end_ev = NULL;
		prev   = 0;
		pos    = score->ch[i].head;
		while (pos > 0) {
			ev = score->events + pos;
			if ((ev->type == TYPE_LABEL) && (ev->len == 0)) {
				track_loop = 1;
			} else if (ev->type == TYPE_END) {
				end_ev = ev;
			}
			if (ev->type == TYPE_LABEL) {
				head[ev->len] = prev;
				scnd[ev->len] = pos;
			} else if (ev->type == TYPE_SKIP) {
				skip[ev->len] = prev;
			} else if (ev->type == TYPE_SKIP_W) {
				skip[ev->len] = prev;
			} else if (ev->type == TYPE_REPEAT) {
				tail[ev->len] = prev;
			} else if (ev->type == TYPE_REPEAT_W) {
				tail[ev->len] = prev;
			} else if (ev->type == TYPE_RET) {
				tail[ev->len] = prev;
			}
			prev = pos;
			pos  = ev->next;
		}
		if (track_loop) {
			if (end_ev != NULL) {
				end_ev->type = TYPE_JMP;
			} else {
				append_event(score, i, TYPE_JMP, 0, 0, 0);
			}
		} else {
			//if ((ev != NULL) && (ev->type != TYPE_JMP) && (ev->type != TYPE_RET) && (ev->type != TYPE_END)) {
			if ((ev != NULL) && (ev->type != TYPE_JMP) && (ev->type != TYPE_END)) {
				append_event(score, i, TYPE_END, 0, 0, 0);
			}
		}

		// XLbv胊s[gϊ
		for (k = 1; tail[k] > 0; k++) {
			if (skip[k] > 0) {
				ev = score->events + tail[k];
				pos = ev->next;
				ev = score->events + pos;
				if ((ev->type == TYPE_REPEAT) || (ev->type == TYPE_REPEAT_W)) {
					// WvCxgǉ
					prev = head[k];
					ev = score->events + prev;
					while (ev->next != scnd[k]) {
						prev = ev->next;
						ev = score->events + prev;
					}
					insert_event(score, i, prev, TYPE_JMP, 0, score->ch[i].label, 0);
					// XLbvCxgxɕύX
					ev = score->events + skip[k];
					pos = ev->next;
					ev = score->events + pos;
					ev->type = TYPE_LABEL;
					ev->len  = score->ch[i].label;
					score->ch[i].label++;
					// s[g擪猳XLbv܂ł
					// XLbv烊s[g܂łւ
					ev = score->events + prev;
					ev = score->events + ev->next;
					ev = score->events + ev->next;
					pos_head = ev->next;
					ev->next = score->events[pos].next;
					ev = score->events + tail[k];
					pos_tail = ev->next;
					ev->next = pos;
					score->events[pos].next = pos_head;
					ev = score->events + skip[k];
					ev->next = pos_tail;
				}
			}
		}
	}

	return 0;
}

int insert_label_event(score_t *score) {
	call_t tbl[LABEL_MAX];
	event_t *ev;
	int pos, ref;
	int i;

	for (i = 0; i < CH_MAX; i++) {
		memset(tbl, 0, sizeof(tbl));

		pos = score->ch[i].head;
		while (pos > 0) {
			ev = score->events + pos;
			if ((ev->type == TYPE_LABEL) && (ev->value > 0)) {
				tbl[ev->value].pos   = pos;
				tbl[ev->value].count = 0;
			}
			pos = ev->next;
		}

		pos = score->ch[i].head;
		while (pos > 0) {
			ev = score->events + pos;
			if (ev->type == TYPE_CALL) {
				ref = tbl[ev->value].pos;
				if (ref > 0) {
					if (tbl[ev->value].count > 0) {
						ev->len = score->ch[i].label;
						insert_event(score, i, ref, TYPE_LABEL, ev->value, ev->len, 0);
						score->ch[i].label++;
					} else {
						ev->len = score->events[ref].len;
					}
					tbl[ev->value].count++;
				}
			}
			pos = ev->next;
		}
	}

	return 0;
}

int insert_tone_event(score_t *score) {
	event_t *ev;
	int pos, prev, sect;
	int low, high;
	int last_low, last_high;
	int i;

	for (i = 0; i < CH_MAX; i++) {
		sect = 0;
		prev = 0;
		pos  = score->ch[i].head;
		low  = 0;
		high = 0;
		last_low  = 0;
		last_high = 0;
		while (pos > 0) {
			ev = score->events + pos;
			if (ev->type == TYPE_NOTE) {
				if (low == 0) {
					sect = prev;
					low  = ev->value;
					high = ev->value;
				} else {
					if (ev->value < low) {
						if (ev->value <= high - TONE_RANGE) {
							insert_event(score, i, sect, TYPE_TONE_BASE, low, high, 0);
							sect = prev;
							low  = ev->value;
							high = ev->value;
						} else {
							low  = ev->value;
						}
					} else if (ev->value > high) {
						if (ev->value >= low + TONE_RANGE) {
							insert_event(score, i, sect, TYPE_TONE_BASE, low, high, 0);
							sect = prev;
							low  = ev->value;
							high = ev->value;
						} else {
							high = ev->value;
						}
					}
				}
			} else if (ev->type == TYPE_LABEL) {
				if (low != 0) {
					insert_event(score, i, sect, TYPE_TONE_BASE, low, high, 0);
					sect = prev;
					last_low  = low;
					last_high = high;
					low  = 0;
					high = 0;
				}
			} else if ((ev->type == TYPE_JMP) || (ev->type == TYPE_CALL)) {
				//if (low == 0) {
				if ((low == 0) && (last_low != 0)) {
					insert_event(score, i, prev, TYPE_TONE_BASE, last_low, last_high, 0);
					sect = pos;
				}
			}
			prev = pos;
			pos  = ev->next;
		}
		if (low > 0) {
			insert_event(score, i, sect, TYPE_TONE_BASE, low, high, 0);
		}
	}

	return 0;
}

int merge_len_table(lentbl_t *tbl1, lentbl_t *tbl2) {
	unsigned long value1, value2;
	int mask, cnt;
	int i, j;
	int r = 0;

	mask = 0;
	cnt  = 0;
	for (i = 0; i < tbl1->count; i++) {
		for (j = 0; j < tbl2->count; j++) {
			if ((tbl1->len[i]   == tbl2->len[j]) &&
			    (tbl1->quant[i] == tbl2->quant[j])) {
				mask |= (1 << j);
				cnt++;
			}
		}
	}
	if (cnt == tbl1->count) {
		tbl1->count = 0;
		r = 1;
	} else if (cnt == tbl2->count) {
		tbl2->count = 0;
		r = 2;
	} else if ((tbl2->count - cnt) <= (LENTBL_SIZE - tbl1->count)) {
		for (j = 0; j < tbl2->count; j++) {
			if ((mask & (1 << j)) == 0) {
				value2 = (tbl2->len[j] << 8) | tbl2->quant[j];
				for (i = tbl1->count; i > 0; --i) {
					value1 = (tbl1->len[i-1] << 8) | tbl1->quant[i-1];
					if (value1 <= value2) break;
					tbl1->len[i]   = tbl1->len[i-1];
					tbl1->quant[i] = tbl1->quant[i-1];
				}
				tbl1->len[i]   = tbl2->len[j];
				tbl1->quant[i] = tbl2->quant[j];
				tbl1->count++;
			}
		}
		tbl2->count = 0;
		r = 2;
	}

	return r;
}

int share_len_table(lentbl_t *tbl1, lentbl_t *tbl2) {
	lentbl_t tmp_tbl;
	unsigned long value1, value2;
	int mask1, mask2, cnt;
	int i, j;
	int r = 0;

	mask1 = 0;
	mask2 = 0;
	cnt  = 0;
	for (i = 0; i < tbl1->count; i++) {
		for (j = 0; j < tbl2->count; j++) {
			if ((tbl1->len[i]   == tbl2->len[j]) &&
			    (tbl1->quant[i] == tbl2->quant[j])) {
				mask1 |= (1 << i);
				mask2 |= (1 << j);
				cnt++;
			}
		}
	}

	j = 0;
	for (i = 0; i < tbl1->count; i++) {
		if (mask1 & (1 << i)) {
			tmp_tbl.len[j]   = tbl1->len[i];
			tmp_tbl.quant[j] = tbl1->quant[i];
			j++;
		}
	}
	tmp_tbl.count = j;

	if (tmp_tbl.count > 0) {
		tbl1->share_cnt = tmp_tbl.count;

		j = 0;
		for (i = 0; i < tbl1->count; i++) {
			if ((mask1 & (1 << i)) == 0) {
				tbl1->len[j]   = tbl1->len[i];
				tbl1->quant[j] = tbl1->quant[i];
				j++;
			}
		}
		for (i = 0; i < tmp_tbl.count; i++) {
			tbl1->len[j]   = tmp_tbl.len[i];
			tbl1->quant[j] = tmp_tbl.quant[i];
			j++;
		}

		j = tbl2->count;
		for (i = tbl2->count-1; i >= 0; --i) {
			if ((mask2 & (1 << i)) == 0) {
				--j;
				tbl2->len[j]   = tbl2->len[i];
				tbl2->quant[j] = tbl2->quant[i];
			}
		}
		for (i = 0; i < tmp_tbl.count; i++) {
			tbl2->len[i]   = tmp_tbl.len[i];
			tbl2->quant[i] = tmp_tbl.quant[i];
		}
	}

	return r;
}

int insert_lentbl_event(score_t *score) {
	lentbl_t *tbl1;
	lentbl_t *tbl2;
	event_t *ev;
	unsigned long value1, value2;
	int pos, prev, sect;
	int i, j, k;

	// e[u̍쐬
	for (i = 0; i < CH_MAX; i++) {
		sect = 0;
		prev = 0;
		pos  = score->ch[i].head;
		tbl1 = NULL;
		while (pos > 0) {
			ev = score->events + pos;
			if ((ev->type == TYPE_NOTE) || (ev->type == TYPE_REST) || (ev->type == TYPE_EXTEND)) {
				if (tbl1 == NULL) {
					sect = prev;
					tbl1 = score->lentbl + score->ltcnt;
					score->ltcnt++;
					tbl1->len[0]   = ev->len;
					tbl1->quant[0] = ev->quant;
					tbl1->count = 1;
					tbl1->share_cnt = 0;
				} else {
					for (j = 0; j < tbl1->count; j++) {
						if ((tbl1->len[j]   == ev->len  ) &&
						    (tbl1->quant[j] == ev->quant)) break;
					}
					if (j >= tbl1->count) {
						if (tbl1->count >= LENTBL_SIZE) {
							insert_event(score, i, sect, TYPE_LEN_TABLE, score->ltcnt-1, 0, 0);
							sect = prev;

							tbl1 = score->lentbl + score->ltcnt;
							score->ltcnt++;
							tbl1->len[0]   = ev->len;
							tbl1->quant[0] = ev->quant;
							tbl1->count = 1;
							tbl1->share_cnt = 0;
						} else {
							value2 = (ev->len << 8) | ev->quant;
							for (j = tbl1->count; j > 0; --j) {
								value1 = (tbl1->len[j-1] << 8) | tbl1->quant[j-1];
								if (value1 <= value2) break;
								tbl1->len[j]   = tbl1->len[j-1];
								tbl1->quant[j] = tbl1->quant[j-1];
							}
							tbl1->len[j]   = ev->len;
							tbl1->quant[j] = ev->quant;
							tbl1->count++;
						}
					}
				}
			} else if (ev->type == TYPE_LABEL) {
				if (tbl1 != NULL) {
					insert_event(score, i, sect, TYPE_LEN_TABLE, score->ltcnt-1, 0, 0);
					sect = prev;
					tbl1 = NULL;
				}
			} else if ((ev->type == TYPE_JMP) || (ev->type == TYPE_CALL)) {
				if ((tbl1 == NULL) && (sect != 0)) {
					insert_event(score, i, prev, TYPE_LEN_TABLE, score->ltcnt-1, 0, 0);
					sect = pos;
				}
/*
				if (tbl1 == NULL) {
					if (sect == 0) {
						tbl1 = score->lentbl + score->ltcnt;
						score->ltcnt++;
						tbl1->count = 0;
						tbl1->share_cnt = 0;
					}
					insert_event(score, i, prev, TYPE_LEN_TABLE, score->ltcnt-1, 0, 0);
					sect = pos;
				}
*/
			}
			prev = pos;
			pos  = ev->next;
		}
		if (tbl1) {
			insert_event(score, i, sect, TYPE_LEN_TABLE, score->ltcnt-1, 0, 0);
		}
	}

	// e[u̓
	for (i = 0; i < score->ltcnt; i++) {
		tbl1 = score->lentbl + i;
		if (tbl1->count > 0) {
			for (j = 0; j < score->ltcnt; j++) {
				tbl2 = score->lentbl + j;
				if ((i != j) && (tbl2->count > 0)) {
					switch (merge_len_table(tbl1, tbl2)) {
					case 1:
						for (k = 0; k < CH_MAX; k++) {
							pos = score->ch[k].head;
							while (pos > 0) {
								ev = score->events + pos;
								if ((ev->type == TYPE_LEN_TABLE) && (ev->value == i)) {
									ev->value = j;
								}
								pos = ev->next;
							}
						}
						break;
					case 2:
						for (k = 0; k < CH_MAX; k++) {
							pos = score->ch[k].head;
							while (pos > 0) {
								ev = score->events + pos;
								if ((ev->type == TYPE_LEN_TABLE) && (ev->value == j)) {
									ev->value = i;
								}
								pos = ev->next;
							}
						}
						break;
					}
				}
			}
		}
	}

	// ׂ荇e[uɋʂ鉹𒊏o
	j = 0;
	for (i = 0; i < score->ltcnt; i++) {
		tbl1 = score->lentbl + i;
		if (tbl1->count > 0) {
			if (j & 1) {
				share_len_table(tbl2, tbl1);
			} else {
				tbl2 = tbl1;
			}
			j++;
		}
	}

	// e[ũx[Xʒuݒ
	pos = (score->rpcnt + 1) / 2;
	for (i = 0; i < score->ltcnt; i++) {
		tbl1 = score->lentbl + i;
		tbl1->base = pos;
		pos += tbl1->count - tbl1->share_cnt;
	}

	return 0;
}

int optimize_event(score_t *score) {
	opttbl_t tbl[LABEL_MAX];
	event_t *ev;
	int pos, prev;
	int label1, label2;
	int low, high, ltbl;
	int i, b, f;

	for (i = 0; i < CH_MAX; i++) {
		for (;;) {
			memset(tbl, 0, sizeof(tbl));

			low    = 0;
			high   = 0;
			ltbl   = -1;
			label1 = -1;
			label2 = -1;
			prev   = -2;
			pos = score->ch[i].head;
			while (pos > 0) {
				ev = score->events + pos;
				if (ev->type == TYPE_LABEL) {
/*
					if (label1 >= 0) {
						tbl[ev->len].link = label1;
					} else {
						tbl[ev->len].link = -1;
					}
*/
					tbl[ev->len].link1 = label1;
					tbl[ev->len].link2 = label2;
					label1 = ev->len;
					label2 = ev->len;
					tbl[label1].near.pos  = prev;
					tbl[label1].near.low  = low;
					tbl[label1].near.high = high;
					tbl[label1].near.ltbl = ltbl;
				} else if (ev->type == TYPE_REPEAT) {
					tbl[ev->len].far.pos  = prev;
					tbl[ev->len].far.low  = low;
					tbl[ev->len].far.high = high;
					tbl[ev->len].far.ltbl = ltbl;
				} else if (ev->type == TYPE_REPEAT_W) {
					tbl[ev->len].far.pos  = prev;
					tbl[ev->len].far.low  = low;
					tbl[ev->len].far.high = high;
					tbl[ev->len].far.ltbl = ltbl;
				} else if (ev->type == TYPE_JMP) {
					tbl[ev->len].far.pos  = prev;
					tbl[ev->len].far.low  = low;
					tbl[ev->len].far.high = high;
					tbl[ev->len].far.ltbl = ltbl;
					low    = 0;
					high   = 0;
					ltbl   = -2;
					label1 = -1;
					label2 = -1;
					prev   = -1;
				} else if (ev->type == TYPE_CALL) {
					tbl[ev->len].far.pos  = prev;
					tbl[ev->len].far.low  = low;
					tbl[ev->len].far.high = high;
					tbl[ev->len].far.ltbl = ltbl;
				} else if (ev->type == TYPE_TONE_BASE) {
					prev = pos;
					low  = ev->value;
					high = ev->len;
				} else if (ev->type == TYPE_LEN_TABLE) {
					ltbl = ev->value;
					label2 = -1;
				} else if (ev->type == TYPE_NOTE) {
					label1 = -1;
					label2 = -1;
				}
				pos = ev->next;
			}

			ltbl  = -1;
			label1 = -1;
			pos = score->ch[i].head;
			while (pos > 0) {
				ev = score->events + pos;
				if (ev->type == TYPE_LABEL) {
					label1 = ev->len;
				} else if (ev->type == TYPE_JMP) {
					ltbl  = -2;
					label1 = -1;
				} else if (ev->type == TYPE_TONE_BASE) {
/*
					if (label1 >= 0) {
						printf("label:%3d (%3d: %2d %2d) (%3d: %2d %2d) (%3d: %2d %2d)\n",
							label1, 
							pos, ev->value, ev->len,
							tbl[label1].near.pos, tbl[label1].near.low, tbl[label1].near.high,
							tbl[label1].far.pos, tbl[label1].far.low, tbl[label1].far.high);
						label2 = tbl[label1].link1;
						while (label2 >= 0) {
							printf("          label:%3d    (%3d: %2d %2d) (%3d: %2d %2d)\n",
								label2, 
								tbl[label2].near.pos, tbl[label2].near.low, tbl[label2].near.high,
								tbl[label2].far.pos, tbl[label2].far.low, tbl[label2].far.high);
							label2 = tbl[label2].link1;
						}
					}
*/
					if (label1 >= 0) {
						low  = ev->value;
						high = ev->len;
						label2 = label1;
						while (label2 >= 0) {
							if ((tbl[label2].near.pos >= 0) && (tbl[label2].link1 < 0)) {
								if (tbl[label2].near.low  < low ) low  = tbl[label2].near.low;
								if (tbl[label2].near.high > high) high = tbl[label2].near.high;
							}
							if (tbl[label2].far.pos >= 0) {
								if (tbl[label2].far.low  < low ) low  = tbl[label2].far.low;
								if (tbl[label2].far.high > high) high = tbl[label2].far.high;
							}
							label2 = tbl[label2].link1;
						}
						if ((high - low + 1) <= TONE_RANGE) {
							b = 0;
							f = 1;
							b |= (ev->value != low)  ? (f<<2) : 0;
							b |= (ev->len   != high) ? (f<<3) : 0;
							f <<= 4;
							ev->value = low;
							ev->len   = high;
							label2 = label1;
							while (label2 >= 0) {
								if (tbl[label2].near.pos == -2) b |= 1;
								if ((tbl[label2].near.pos >= 0) && (tbl[label2].link1 < 0)) {
									b |= (score->events[tbl[label2].near.pos].value != low)  ? (f   ) : 0;
									b |= (score->events[tbl[label2].near.pos].len   != high) ? (f<<1) : 0;
									score->events[tbl[label2].near.pos].value = low;
									score->events[tbl[label2].near.pos].len   = high;
								}
								if (tbl[label2].far.pos == -2) b |= 1;
								if (tbl[label2].far.pos >= 0) {
									b |= (score->events[tbl[label2].far.pos].value != low)  ? (f<<2) : 0;
									b |= (score->events[tbl[label2].far.pos].len   != high) ? (f<<3) : 0;
									score->events[tbl[label2].far.pos].value = low;
									score->events[tbl[label2].far.pos].len   = high;
								}
								f <<= 4;
								label2 = tbl[label2].link1;
							}
							ev->quant = (b & 1) ? 0 : 1;
							if (b & -4) {
//								printf("----(%02x)\n", b);
								break;
							}
						} else {
							ev->quant = 0;
						}
					}
					low  = ev->value;
					high = ev->len;
				} else if (ev->type == TYPE_LEN_TABLE) {
/*
					if (label1 >= 0) {
						printf("label:%3d ltbl:%3d ltbl:%3d ltbl:%3d\n",
							label1, ev->value, ltbl, tbl[label1].ltbl);
					}
*/
					if ((label1 >= 0) && (ev->quant == 0)) {
						f = 1;
//						b = (ev->value != ltbl) ? f : 0;
						b = ((ltbl >= -1) && (ev->value != ltbl)) ? f : 0;
						f <<= 1;
						label2 = label1;
/*
						while (label2 >= 0) {
							b |= (ev->value != tbl[label2].near.ltbl) ? (f   ) : 0;
							b |= (ev->value != tbl[label2].far.ltbl)  ? (f<<1) : 0;
							f <<= 2;
							label2 = tbl[label2].link2;
						}
*/
						while (label2 >= 0) {
							b |= ((tbl[label2].near.ltbl >= -1) && (ev->value != tbl[label2].near.ltbl)) ? (f   ) : 0;
							b |= ((tbl[label2].far.ltbl  >= -1) && (ev->value != tbl[label2].far.ltbl))  ? (f<<1) : 0;
							f <<= 2;
							label2 = tbl[label2].link2;
						}
						ev->quant = (b) ? 0 : 1;
					}
					ltbl = ev->value;
				} else if (ev->type == TYPE_NOTE) {
					label1 = -1;
				}
				pos = ev->next;
			}

			if (pos == 0) break;
		}
	}

	return 0;
}

int assign_label_offset(score_t *score) {
	int locate[LABEL_MAX];
	event_t *ev;
	int pos, cnt;
	int i, j, n;

	for (i = 0; i < CH_MAX; i++) {
		memset(locate, 0, sizeof(locate));

		for (j = 0; j < 3; j++) {
			cnt = 0;
			pos  = score->ch[i].head;
			while (pos > 0) {
				ev = score->events + pos;
				if (ev->type == TYPE_NOTE) {
					cnt++;
				} else if (ev->type == TYPE_REST) {
					cnt++;
				} else if (ev->type == TYPE_EXTEND) {
					cnt++;
				} else if (ev->type == TYPE_REG1) {
					cnt += 2;
				} else if (ev->type == TYPE_REG2) {
					cnt += 2;
				} else if ((ev->type == TYPE_TONE_BASE) && (ev->quant == 0)) {
					cnt += 2;
				} else if ((ev->type == TYPE_LEN_TABLE) && (ev->quant == 0)) {
					cnt += 2;
				} else if (ev->type == TYPE_ENV_BASE) {
					cnt += 2;
				} else if ((ev->type == TYPE_TEMPO) && (ev->quant == 0)) {
					cnt += 2;
				} else if (ev->type == TYPE_LOOP_CNT) {
					cnt++;
				} else if (ev->type == TYPE_REPEAT) {
					n = locate[ev->len] - (cnt+1);
					if ((ev->value <= 9) && (n > -32)) {
						cnt += 2;
						ev->quant = 1;
						n = locate[ev->len] - cnt;
					} else {
						cnt += 3;
						ev->quant = 2;
						n = locate[ev->len] - cnt;
					}
					if (j == 2) {
						if ((ev->value <= 127) && (n > -256)) {
							ev->len = (locate[ev->len] - cnt) & 0xff;
						} else {
							if (ev->value > 255) {
								printf("repeat count error.\n");
							}
							if (n <= -256) {
								printf("repeat distance error.\n");
							}
						}
					}
				} else if (ev->type == TYPE_REPEAT_W) {
					n = locate[ev->len] - (cnt+1);
					if ((ev->value <= 9) && (n > -32)) {
						cnt += 2;
						ev->quant = 1;
						n = locate[ev->len] - cnt;
					} else {
						cnt += 3;
						ev->quant = 2;
						n = locate[ev->len] - cnt;
					}
					if (j == 2) {
						if ((ev->value <= 127) && (n > -256)) {
							ev->len = (locate[ev->len] - cnt) & 0xff;
						} else {
							if (ev->value > 255) {
								printf("repeat count error.\n");
							}
							if (n <= -256) {
								printf("repeat distance error.\n");
							}
						}
					}
				} else if ((ev->type == TYPE_JMP) || (ev->type == TYPE_CALL)) {
					if (j == 0) {
						cnt += 3;
					} else if (j == 1) {
						n = locate[ev->len] - (cnt+2);
						if ((n > -64) && (n < 64)) {
							cnt += 2;
						} else {
							cnt += 3;
						}
					} else {
						n = locate[ev->len] - (cnt+2);
						if ((n > -64) && (n < 64)) {
							ev->quant = 0;
							cnt += 2;
						} else {
							ev->quant = 1;
							cnt += 3;
						}
						ev->len = (locate[ev->len] - cnt) & 0xffff;
					}
				} else if (ev->type == TYPE_RET) {
					cnt += 1;
				} else if (ev->type == TYPE_END) {
					cnt += 1;
				} else if (ev->type == TYPE_LABEL) {
					if (j < 2) {
						locate[ev->len] = cnt;
					}
				}
				pos  = ev->next;
			}
		}
	}

	return 0;
}

/*
int assign_envtable_base(score_t *score) {
	int count;
	int i, n;

	count = 2;
	for (i = 0; i < ENV_MAX; i++) {
		score->env[i].base = count;
		count += score->env[i].size;
	}

	for (i = 0; i < score->rpcnt; i++) {
		n = score->rhythm[i].data & 15;
		if (n > 0) {
			n = score->env[n-1].base;
			if (n <= 15) {
				score->rhythm[i].data &= 0xfff0;
				score->rhythm[i].data |= n;
			} else {
				printf("rhythm envelope error. (%s)\n", score->rhythm[i].name);
			}
		}
	}

	return 0;
}
*/

int write_cmd(score_t *score, FILE *fp, const char *prefix) {
	int valid;
	int i;

	valid = (score->tempo_1st & 255) << 8;
	for (i = 0; i < CH_MAX; i++) {
		if (score->ch[i].head > 0) {
			valid |= (1 << i);
		}
	}

	fprintf(fp, "%s_PlayData:\n", prefix);
/*
	fprintf(fp, "\t.dw\t$%04x, %s_LenTable-2, %s_EnvTable\n", valid, prefix, prefix);
*/
	fprintf(fp, "\t.dw\t$%04x, %s_LenTable-2\n", valid, prefix);
	fprintf(fp, "\t.dw\t", prefix);
	for (i = 0; i < CH_MAX; i++) {
		if (valid & (1 << i)) {
			fprintf(fp, "%s_Seq%c", prefix, i+'A');
			if (((valid & 15) & ~((1<<(i+1))-1)) != 0) {
				fprintf(fp, ", ");
			}
		}
	}
	fprintf(fp, "\n");

	return 0;
}

int write_data(score_t *score, FILE *fp, const char *prefix) {
	unsigned char buf[4];
	event_t *ev;
	lentbl_t *tbl;
	int pos, cnt;
	int tone_base;
	int tick_min;
	int i, j, n;

	// e[u
	fprintf(fp, "%s_LenTable:\n", prefix);
	cnt = 0;
	for (i = 0; i < score->rpcnt; i++) {
		if ((cnt % 8) == 0) {
			fprintf(fp, "\t.db\t");
		} else {
			fprintf(fp, ", ");
		}
		fprintf(fp, "$%02x", score->rhythm[i].data);
		if ((cnt % 8) == 7) {
			fprintf(fp, "\n");
		}
		cnt++;
	}
	if ((score->rpcnt & 1) != 0) {
		fprintf(fp, ", $00");
		if ((cnt % 8) == 7) {
			fprintf(fp, "\n");
		}
		cnt++;
	}
	if ((cnt % 8) != 0) {
		fprintf(fp, "\n");
	}
	tick_min = (score->tempo_max+75-1) / 75;
	cnt = 0;
	for (i = 0; i < score->ltcnt; i++) {
		tbl = score->lentbl + i;
		for (j = 0; j < (tbl->count - tbl->share_cnt); j++) {
			if ((cnt % 4) == 0) {
				fprintf(fp, "\t.db\t");
			} else {
				fprintf(fp, ", ");
			}
			if ((tbl->quant[j] == 8) || (tbl->len[j] <= tick_min)) {
				n = 0;
			} else {
				n = tbl->len[j] * (8 - tbl->quant[j]) / 8;
				if (n < tick_min) {
					if (tbl->len[j] - tick_min < tick_min) {
						n = 0;
					} else {
						n = tick_min;
					}
				}
			}
			fprintf(fp, "$%02x, $%02x", (-n)&255, (-(tbl->len[j]-n))&255);
			if ((cnt % 4) == 3) {
				fprintf(fp, "\n");
			}
			cnt++;
		}
	}
	if ((cnt % 4) != 0) {
		fprintf(fp, "\n");
	}

	// V[PXf[^
	for (i = 0; i < CH_MAX; i++) {
		if (score->ch[i].head > 0) {
			fprintf(fp, "%s_Seq%c:\n", prefix, i+'A');
			pos = score->ch[i].head;
			cnt = 0;
			tone_base = 0;
			tbl = NULL;
			while (pos > 0) {
				ev = score->events + pos;
				switch (ev->type) {
				case TYPE_NOTE:
					if (tbl) {
						for (j = 0; j < LENTBL_SIZE; j++) {
							if ((tbl->len[j] == ev->len) && (tbl->quant[j] == ev->quant)) break;
						}
						if (j >= LENTBL_SIZE) j = 0;
					} else {
						j = 0;
					}
					buf[0] = ((j+1) << 4) | (ev->value - tone_base + 2);
					n = 1;
					break;
				case TYPE_REST:
					if (tbl) {
						for (j = 0; j < LENTBL_SIZE; j++) {
							if ((tbl->len[j] == ev->len) && (tbl->quant[j] == ev->quant)) break;
						}
						if (j >= LENTBL_SIZE) j = 0;
					} else {
						j = 0;
					}
					buf[0] = ((j+1) << 4) | 0x01;
					n = 1;
					break;
				case TYPE_EXTEND:
					if (tbl) {
						for (j = 0; j < LENTBL_SIZE; j++) {
							if ((tbl->len[j] == ev->len) && (tbl->quant[j] == ev->quant)) break;
						}
						if (j >= LENTBL_SIZE) j = 0;
					} else {
						j = 0;
					}
					buf[0] = ((j+1) << 4);
					n = 1;
					break;
				case TYPE_REG1:
					buf[0] = 0x08;
					buf[1] = ev->value;
					n = 2;
					break;
				case TYPE_REG2:
					buf[0] = 0x09;
					buf[1] = ev->value;
					n = 2;
					break;
				case TYPE_TONE_BASE:
					if (ev->quant == 0) {
						buf[0] = 0x0a;
						buf[1] = ev->value - 2*12;
						n = 2;
					} else {
						n = 0;
					}
					tone_base = ev->value;
					break;
				case TYPE_LEN_TABLE:
					if (ev->quant == 0) {
						buf[0] = 0x0b;
						buf[1] = score->lentbl[ev->value].base;
						n = 2;
					} else {
						n = 0;
					}
					tbl = score->lentbl + ev->value;
					break;
				case TYPE_ENV_BASE:
					buf[0] = 0x0c;
					buf[1] = (ev->value > 0) ? score->env[ev->value-1].value : 0;
					n = 2;
					break;
				case TYPE_TEMPO:
					if (ev->quant == 0) {
						buf[0] = 0x0e;
						buf[1] = ev->value;
						n = 2;
					} else {
						n = 0;
					}
					break;
				case TYPE_LOOP_CNT:
					buf[0] = 0x0f;
					n = 1;
					break;
				case TYPE_CALL:
					if (ev->quant == 0) {
						buf[0] = 0x02;
						buf[1] = ev->len & 0x7f;
						n = 2;
					} else {
						buf[0] = 0x02;
						buf[1] = ((ev->len   )&0x7f)|0x80;
						buf[2] = ((ev->len>>7)&0xff);
						n = 3;
					}
					break;
				case TYPE_RET:
					buf[0] = 0x01;
					n = 1;
					break;
				case TYPE_REPEAT:
					if (ev->quant == 1) {
						buf[0] = 0x04;
						buf[1] = (((ev->value-2)<<5)|(ev->len & 0x1f)) & 0xff;
						n = 2;
					} else {
						buf[0] = 0x05;
						buf[1] = ev->value & 0xff;
						buf[2] = ev->len & 0xff;
						n = 3;
					}
					break;
				case TYPE_REPEAT_W:
					if (ev->quant == 1) {
						buf[0] = 0x06;
						buf[1] = (((ev->value-2)<<5)|(ev->len & 0x1f)) & 0xff;
						n = 2;
					} else {
						buf[0] = 0x07;
						buf[1] = ev->value & 0xff;
						buf[2] = ev->len & 0xff;
						n = 3;
					}
					break;
				case TYPE_JMP:
					if (ev->quant == 0) {
						buf[0] = 0x03;
						buf[1] = ev->len & 0x7f;
						n = 2;
					} else {
						buf[0] = 0x03;
						buf[1] = ((ev->len   )&0x7f)|0x80;
						buf[2] = ((ev->len>>7)&0xff);
						n = 3;
					}
					break;
				case TYPE_END:
					buf[0] = 0x00;
					n = 1;
					break;
				default:
					n = 0;
				}
				for (j = 0; j < n; j++) {
					if ((cnt % 8) == 0) {
						fprintf(fp, "\t.db\t");
					} else {
						fprintf(fp, ", ");
					}
					fprintf(fp, "$%02x", buf[j]);
					if ((cnt % 8) == 7) {
						fprintf(fp, "\n");
					}
					cnt++;
				}
				pos = ev->next;
			}
			if ((cnt % 8) != 0) {
				fprintf(fp, "\n");
			}
		}
	}

	return 0;
}

void dump_event(event_t *events, int pos) {
	static const char *note[] = { "c", "c+", "d", "d+", "e", "f", "f+", "g", "g+", "a", "a+", "b" };
	event_t *ev;
	int tone;
	int i;

	while (pos > 0) {
		ev = events + pos;
		printf("type:%2d value:%5d len:%5d quant:%2d", ev->type, ev->value, ev->len, ev->quant);
		if (ev->type == TYPE_NOTE) {
			tone = ev->value;
			printf(" (o%d%s)", tone/12, note[tone%12]);
		} else if (ev->type == TYPE_REST) {
			printf(" (r)");
		} else if (ev->type == TYPE_EXTEND) {
			printf(" (x)");
		} else if (ev->type == TYPE_REG1) {
			printf(" (reg1)");
		} else if (ev->type == TYPE_REG2) {
			printf(" (reg2)");
		}
		printf("\n");
		pos = ev->next;
	}
}

void dump_lentbl(lentbl_t *lentbl) {
	int i;

	printf("[ ");
	for (i = 0; i < lentbl->count; i++) {
		printf("%d:%d ", lentbl->len[i], lentbl->quant[i]);
	}
	printf("]\n");
}

void dump(score_t *score) {
	int i;

	for (i = 0; i < CH_MAX; i++) {
		printf("==== ch%c ====\n", i+'A');
		dump_event(score->events, score->ch[i].head);
	}
	for (i = 0; i < score->ltcnt; i++) {
		printf("==== lentbl%d ====\n", i);
		dump_lentbl(score->lentbl + i);
	}
}
