# Fssq (Famicom Sound Sequencer) 仕様書

## 概要

Fssqlは、ファミコン/NES用の音楽ドライバです。
下記の機能をサポートしています。

* mapper
    * NROM (mapper0) に対応しています。
* BGM演奏
    * 曲データを元に
      2A03のpAPUの各音源(パルス波x2、三角波、ノイズ、DPCM)を
      制御して曲を演奏します。
* SE演奏
    * BGM演奏中にSE用の曲データを割りこませて演奏します。
    * 2A03のpAPUの各音源の内、DPCMを除いた音源を１つまで
      使用して演奏することが可能です。
* BGM演奏一時中断
    * BGM演奏を一時的に止めて再開することが可能です。
* BGM演奏フェードアウト
    * BGMの音量を徐々に小さくした後、演奏を停止することが可能です。

## 仕様

### 音長

本音楽ドライバでは、全音符の長さを192とするtickで
音長を管理します。

tickは整数で管理するため、割り切れない音長は利用できません。
(例 128b分音符=1.5tick など)

音長を保持するメモリ領域は8bitで全音符の長さの192tickまでを管理します。<br>
全音符を越える長さの音符/休符は、コンパイラにより音符/休符コマンドと
延長コマンドに分解されます。<br>
例えば、クォンタイズが7で480tick(全音符＋全音符＋二分音符)の音符は
```
音符コマンド(キーオン 96tick, キーオフ  0tick)
延長コマンド(キーオン192tick, キーオフ  0tick)
延長コマンド(キーオン168tick, キーオフ 24tick)
```
の３コマンドに分解されます。<br>
クォンタイズの指定は分割したコマンドの末尾のみ適用されます。

### チャンネル

本音楽ドライバでは、pAPUの各音源をチャンネルと呼びます。

|   |   |
|---|---|
|矩形波1|Aチャンネル|
|矩形波2|Bチャンネル|
|三角波 |Cチャンネル|
|ノイズ |Dチャンネル|

pAPUには上記の音源以外にDPCMが存在しますが、
本音楽ドライバでは独立したチャンネルとしては扱わず、
後述する特殊コマンドを使用して発音、制御します。

### トラック

本音楽ドライバでは、各チャンネル別のシーケンスデータの集まりを
トラックと呼びます。<br>
BGM演奏の場合には４トラック(内蔵音源のみの場合)まで使用できます。<br>
SE演奏の場合には１トラックまで使用できます。


## コンパイラ

mml2fssqlは、テキストファイルに記述したMMLを
音楽ドライバ(sound.asm)が解釈可能なバイナリデータに変換する
コンパイラです。

このコンパイラは、nesasmでアセンブル可能なソースコードの
形式で次の２つのファイルを生成します。

* MLSDファイル
    * 曲データ本体のバイナリデータを定義するファイルです。
    * 曲データ本体は、下記の複数のデータテーブルから構成されます。
        * 音長テーブル
        * シーケンスデータ
* MLSCファイル
    * 下記の情報をまとめたテーブルを定義するファイルです。
        * 使用チャンネル情報
        * テンポ初期値
        * 曲データの各テーブルの先頭ラベル(アドレス)

MLSCファイルに定義されたテーブルの先頭アドレスを
Fssqlの演奏開始サブルーチンに渡すことで、曲が演奏されます。


mml2fssqlはコマンドプロンプトから実行します。
コマンド実行時のオプション指定は次の通りです。

```
mml2fssql <MMLファイル> <MLSCファイル> <MLSDファイル> <ラベル>
```

* <MMLファイル>
    * MMLを記述したテキストファイルのファイル名を指定します。
* <MLSCファイル>
    * MLSCファイルのファイル名を指定します。
* <MLSDファイル>
    * MLSDファイルのファイル名を指定します。
* <ラベル>
    * MLSCファイル、およびMLSDファイルに定義される各テーブルの
      先頭ラベルのプレフィクスを指定します。
      下記のラベルが生成されます。
        * <ラベル>_PlayData
        * <ラベル>_LenTable
        * <ラベル>_EnvTable
        * <ラベル>_SeqA, <ラベル>_SeqB, <ラベル>_SeqC, ...


## サブルーチン

音楽ドライバ(sound.asm)には、APIとして次のサブルーチンが
定義されています。

### Fssql_Init

音楽ドライバを初期化します。

|   |   |
|---|---|
|引数レジスタ  |なし|
|使用レジスタ  |A X|
|使用ゼロページ|なし|


### Fssql_PlayBGM

BGMを演奏します。
曲データで定義された～_PlayDataのラベルのアドレスを引数で指定します。

|   |   |
|---|---|
|引数レジスタ  |A : 曲データのアドレス(LOW)<br>X : 曲データのアドレス(HIGH)|
|使用レジスタ  |A X Y|
|使用ゼロページ|$00 ～ $03|

### Fssql_PlaySE

SEを演奏します。
曲データで定義された～_PlayDataのラベルのアドレスを引数で指定します。

|   |   |
|---|---|
|引数レジスタ  |A : 曲データのアドレス(LOW)<br>X : 曲データのアドレス(HIGH)|
|使用レジスタ  |A X Y|
|使用ゼロページ|$00 ～ $03|

### Fssql_StopBGM

BGMの演奏を停止します。

|   |   |
|---|---|
|引数レジスタ  |なし|
|使用レジスタ  |A|
|使用ゼロページ|なし|

### Fssql_StopSE

SEの演奏を停止します。

|   |   |
|---|---|
|引数レジスタ  |なし|
|使用レジスタ  |A X Y|
|使用ゼロページ|$00 ～ $01|

### Fssql_PauseBGM

BGMの演奏を一時停止します。

|   |   |
|---|---|
|引数レジスタ  |なし|
|使用レジスタ  |A|
|使用ゼロページ|なし|

### Fssql_ResumeBGM

一時停止したBGMの演奏を再開します。

|   |   |
|---|---|
|引数レジスタ  |なし|
|使用レジスタ  |A|
|使用ゼロページ|なし|

### Fssql_Fadeout

演奏しているBGMの音量を徐々に小さくした後、BGMの演奏を停止します。

|   |   |
|---|---|
|引数レジスタ  |なし|
|使用レジスタ  |A|
|使用ゼロページ|なし|

### Fssql_Update

音楽ドライバの演奏処理を行います。
1/60秒毎にこのサブルーチンを呼び出してください。
音源のレジスタへの書き込みは本サブルーチンでは行わないため、
呼び出しタイミングは揺らぎがあっても問題ありません。
本サブルーチンの処理は割り込みがあっても問題ないため、
割り込み許可状態で呼びだすことを推奨します。

|   |   |
|---|---|
|引数レジスタ  |なし|
|使用レジスタ  |A X Y|
|使用ゼロページ|$00 ～ $0b|

### Fssql_Reflect

音楽ドライバが処理したデータを音源のレジスタに書き込みます。
1/60秒毎にこのサブルーチンを呼び出してください。
呼び出しタイミングはなるべく揺らぎがない方が正確な演奏となります。
NMI割り込みの延長で呼びだすことを推奨します。
本サブルーチンの処理は割り込みがあっても問題ないため、
割り込み許可状態で呼びだすことを推奨します。

|   |   |
|---|---|
|引数レジスタ  |なし|
|使用レジスタ  |A X Y|
|使用ゼロページ|なし|

## MML

コンパイラがサポートしているMML表記の詳細を以下に示します。

### データ定義

#### 書式

基本的な書式は次の通りです。

```
[ <定義名> [ <定義データ>... ] ] [ ; <コメント> ]
```

#### 定義名

記述可能な定義を以下に示します。

|定義名|説明|
|---|---|
|#EP|エンベロープデータ定義<br>発音中の音の音量や音程を変化させるエンベロープデータを定義します。|
|#RP|リズム定義<br>!コマンドで使用するリズムを定義します。

#### エンベロープデータ

エンベロープデータは、発音中の音量や音程の変化を付けるデータ定義です。<br>
時間の経過と共に音量や音程が変化することにより、音の聞こえ方、音色が変わります。

エンベロープデータ定義の書式は次の通りです。

```
#EP <番号> <エンベロープデータ>
```

* <番号>
    * エンベロープデータを識別する番号で1～15の値を指定します。
* <エンベロープデータ>
    * 8bitの値で各bitに次のような意味を持ちます。

|ビット|意味|
|---|---|
|7～6bit|音の鳴り始めの数フレームの間矩形波のデューティ比を変化させます。<br>音量変化パターンの指定が0または2の時のみ有効です。<br>@xコマンドで指定した値の上位2bitに本指定値をxorした値が2A03のpAPUレジスタに設定されます。|
|5～4bit|未使用|
|3～2bit|音量変化パターンを指定します。(下記図形を参照)<br>3は指定禁止です。 |
|1～0bit|周波数増加値<br>発音時に音程から表引きした周波数に0～3の値を加算します。|

<svg x=0 y=0 width=200 height=150>
  <text x="10" y="25" storke="none" fill="black">音量変化パターン 0</text>
  <line x1="40" y1="120" x2="150" y2="120" stroke="black" stroke-width="1" />
  <line x1="150" y1="120" x2="190" y2="120" stroke="black" stroke-width="1" />
  <line x1="40" y1="110" x2="40" y2="125" stroke="black" stroke-width="1" />
  <line x1="150" y1="110" x2="150" y2="125" stroke="black" stroke-width="1" />
  <line x1="190" y1="110" x2="190" y2="125" stroke="black" stroke-width="1" />
  <text x="5" y="80" storke="none" fill="black">音量</text>
  <text x="60" y="140" storke="none" fill="black">Key-On</text>
  <text x="145" y="140" storke="none" fill="black">Key-Off</text>
  <line x1="40" y1="80" x2="60" y2="60" stroke="black" stroke-width="1" />
  <line x1="60" y1="60" x2="150" y2="60" stroke="black" stroke-width="1" />
  <line x1="150" y1="60" x2="190" y2="100" stroke="black" stroke-width="1" />
</svg>

<svg x=0 y=0 width=200 height=150>
  <text x="10" y="25" storke="none" fill="black">音量変化パターン 1</text>
  <line x1="40" y1="120" x2="150" y2="120" stroke="black" stroke-width="1" />
  <line x1="150" y1="120" x2="190" y2="120" stroke="black" stroke-width="1" />
  <line x1="40" y1="110" x2="40" y2="125" stroke="black" stroke-width="1" />
  <line x1="150" y1="110" x2="150" y2="125" stroke="black" stroke-width="1" />
  <line x1="190" y1="110" x2="190" y2="125" stroke="black" stroke-width="1" />
  <text x="5" y="80" storke="none" fill="black">音量</text>
  <text x="60" y="140" storke="none" fill="black">Key-On</text>
  <text x="145" y="140" storke="none" fill="black">Key-Off</text>
  <line x1="40" y1="60" x2="60" y2="60" stroke="black" stroke-width="1" />
  <line x1="60" y1="60" x2="150" y2="60" stroke="black" stroke-width="1" />
  <line x1="150" y1="60" x2="190" y2="100" stroke="black" stroke-width="1" />
</svg>

<svg x=0 y=0 width=200 height=150>
  <text x="10" y="25" storke="none" fill="black">音量変化パターン 2</text>
  <line x1="40" y1="120" x2="150" y2="120" stroke="black" stroke-width="1" />
  <line x1="150" y1="120" x2="190" y2="120" stroke="black" stroke-width="1" />
  <line x1="40" y1="110" x2="40" y2="125" stroke="black" stroke-width="1" />
  <line x1="150" y1="110" x2="150" y2="125" stroke="black" stroke-width="1" />
  <line x1="190" y1="110" x2="190" y2="125" stroke="black" stroke-width="1" />
  <text x="5" y="80" storke="none" fill="black">音量</text>
  <text x="60" y="140" storke="none" fill="black">Key-On</text>
  <text x="145" y="140" storke="none" fill="black">Key-Off</text>
  <line x1="40" y1="40" x2="60" y2="60" stroke="black" stroke-width="1" />
  <line x1="60" y1="60" x2="150" y2="60" stroke="black" stroke-width="1" />
  <line x1="150" y1="60" x2="190" y2="100" stroke="black" stroke-width="1" />
</svg>

エンベロープデータの定義例を次に示します。

```
#EP 1 $04
#EP 2 $08
#EP 3 $80
```

#### リズム

リズムデータは、Dチャンネルのシーケンスで、
定型のノイズやDPCMの発音を簡素に記述出来るよう
ノイズの音程、音量、エンベロープ指定、DPCMの打音指定の
組み合わせを定義するデータです。

リズム定義の書式は次の通りです。

```
#RP <名前> <打音番号> <音程> <音量> [ <クォンタイズ> ]
```

* <名前>
    * リズム定義データの名前を指定します。<br>アルファベット(A～Z a～z)とアンダースコア(_)の文字が使用できます。<br>大文字と小文字の区別は行いません。
* <打音番号>
    * DPCMで発音する音の番号を指定します。0～3の番号が指定できます。<br>1はバスドラム、2はスネア、3はタムです。0を指定するとDPCMの発音は行いません。
* <音程>
    * ノイズの音程を指定します。0,1,6,7の値を指定できます。<br>0を指定すると発音しません。
* <音量>
    * ノイズの音量を指定します。0～15の値を指定できます。
* <クォンタイズ>
    * ノイズのキーオン/キーオフの時間比を設定します。1～8の値が指定可能です。<br>値の詳細は後述のシーケンスのコマンドのクォンタイズの説明を参照下さい。<br>記述を省略した場合はMMLコマンド記述中のクォンタイズの指定値が適用されます。

リズムの定義例を次に示します。

```
;   name dpcm tone vol quant
#RP b       1  $00   0       ; kick
#RP h       0  $01   6       ;       hi-hat
#RP bh      1  $01   6       ; kick  hi-hat
#RP hc      0  $01   4     3 ;       hi-hat close
#RP bhc     1  $01   4     3 ; kick  hi-hat close
#RP s       2  $07   4       ;       snare
#RP bs      1  $07   4       ; kick  snare
```

### シーケンス定義

#### 書式

基本的な書式は次の通りです。

```
[ <チャンネル> <MMLコマンド>... ] [ ; <コメント> ]
```

<チャンネル> にMMLの記述対象となるチャンネルを
アルファベットで指定します。

|   |   |
|---|---|
|A|矩形波1|
|B|矩形波2|
|C|三角波|
|D|ノイズ|

<MMLコマンド> にMMLコマンドを列記します。<br>
音楽ドライバは記述した順にMMLコマンドを実行し、音源を制御します。

; 以降はコメントとして無視されます。
コメントのみの行や空行の記述も可能です。

#### コマンド

記述可能なMMLコマンドを以下に示します。

|   |   |
|---|---|
|cdefgab|音符<br>コマンドの直後に+を付けると半音上がり-を付けると半音下がります。<br>コマンドの後に音長を指定することができます。<br>音長を指定しなかった場合にはlコマンドで指定したデフォルトの音長が適用されます。|
|r|休符<br>コマンドの後に音長を指定することができます。<br>音長を指定しなかった場合にはlコマンドで指定したデフォルトの音長が適用されます。|
|!|リズム<br>リズム定義に従ってノイズとDPCMを発音します。<br>コマンド直後にリズム定義(#RP)の最初のパラメータのリズム名を記述します。<br>リズム名の後に音長を指定することができます。<br>Dチャンネルにのみ記述が可能です。
|&|タイ<br>先に発音した音符の発音時間を延長します。<br>コマンドの後に音長を指定することができます。<br>音長を指定しなかった場合にはlコマンドで指定したデフォルトの音長が適用されます。|
|m|参照演奏<br>本コマンドの直後に記述した番号と一致する()コマンドで範囲指定したデータを演奏します。<br>BGMでのみ使用できます。|
|o|オクターブ指定<br>コマンドの直後に記述した番号のオクターブを指定します。<br>番号は2～7の値が指定できます。|
|>|オクターブ上昇<br>現在のオクターブ指定を１つアップします。|
|<|オクターブ下降<br>現在のオクターブ指定を１つダウンします。|
|l|デフォルト音長の設定<br>音符/休符で音長を指定しなかった時に適用する音長を設定します。<br>コマンドの直後に音長を記述します。|
|q|クォンタイズ<br>キーオン/キーオフの時間比を設定します。<br>コマンドの直後に 1～8 の値を指定します。<br>指定した値をnとした場合、キーオンの時間(GateTime)は音長(StepTime)のn/8、<br>キーオフの時間(StepTime-GateTime)は音長の(8-n)/8となります。<br>8を指定した場合、音長の全時間がキーオンの状態となります。|
|t|テンポ<br>テンポを指定します。<br>コマンドの直後に 1～255 の値を指定します。<br>Aチャンネルに指定してください。|
|@x|レジスタ1設定<br>内蔵音源は$4000,$4004,$400cのレジスタに値を設定します。<br>コマンド直後にレジスタに設定する値を指定します。<br><br>矩形波およびノイズではbit5(キーオフカウンタフラグ)、bit4(エンベロープループカウンタ)に1を設定して機能を無効にしてください。<br>三角波では指定しても無視されます。|
|@y|レジスタ2設定<br>内蔵音源は$4001,$4005のレジスタに値を設定します。<br>コマンド直後にレジスタに設定する値を指定します。<br><br>矩形波では、ハードウェアのスイープ機能を使用せず広域リピート機能を使用する場合は$00を設定してください。<br>ハードウェアのスイープ機能を使用せずo2g+以下の低音を使用する場合は$08を設定します。<br>三角波およびノイズではハードウェア的に未使用レジスタであるため、設定しても無意味です。<br>値の保存領域は広域リピート機能のカウンタとして利用するので本コマンドは使用しないでください。|
|@ev|エンベロープ設定<br>適用するエンベロープデータを設定します。<br>コマンド直後にエンベロープデータの番号を指定します。<br>番号に0を指定すると適用なしとなります。|
|@lc|ループカウンタ更新<br>曲のループ回数を手動更新します。<br>仕様上チャンネル毎にループタイミングを変えることが可能で、どのタイミングを曲のループとするか定義できず、<br>自動検出が困難なため、ユーザがタイミングを指示する方式としています。|
|[ \| ]|リピート<br>[ と ] の間に挟まれたMMLを繰り返し演奏します。<br>]の直後に繰り返し回数を指定します。<br>[ と ] の間に \| が記述されていると、最後の繰り返しでは、\| と ] の間のMMLをスキップします。|
|@[ @\| @]|広域リピート<br>@[ と @] の間に挟まれたMMLを繰り返し演奏します。<br>@]の直後に繰り返し回数を指定します。<br>@[ と @] の間に @\| が記述されていると、最後の繰り返しでは、@\| と @] の間のMMLをスキップします。<br><br>カウンタの保存領域を@yコマンドで設定するレジスタ値と共有しているため、<br>内蔵音源の矩形波で使用する場合には次の制限があります。<br>・ハードウェアのスイープ機能を使用する場合には広域リピート機能を使用できません。<br>・o2g+以下の低音を発音する場合には広域リピート機能を使用できません。<br>・広域リピート機能を使用する場合には@yコマンドで$00を設定し、カウンタを初期化します。|
|( )|参照範囲指定<br>mコマンドで参照するコマンドの範囲を指定します。<br>)の直後に識別の番号を指定します。|
|:|トラックループ<br>このコマンドが記述されているとデータ終端まで演奏した後、このコマンドの位置に戻ります。<br>このコマンドが記述されていない場合は、データ終端まで演奏した後、演奏を停止します。|
|/|データ終端<br>通常は各トラックのデータの入力末尾がデータ終端として自動的に認識しますが、<br>参照範囲指定するデータをトラックの末尾に集約して、<br>集約したデータの手前でトラックループまたは演奏停止を行いたい場合に<br>このコマンドを記述します。|


### MMLのサンプル

MMLの記述例を以下に示します。

```
; 3人の王の行列

#EP 1 $08

A t150 @x$b6 @y$08

A o4 q7 @ev1
A :
A ( r16 f16 f4 c4 
A f4.. g16 g+8. g16 g+8. f16 
A > c4 < r8 g+8 a+4 > c4 <
A > c+8 c8 < a+8 g+8 )1 g4 a+4
A a+8 g+8 g8
A m1
A g+4 g4 f4.

A ( r16 g16 g4 g4 
A g+4 g8. f16 g4 g+4
A a+4. r16 g+16 a+4 > c4 < )2
A f4 g8 g+8 a+8 g+8 g8 f8
A f8 e8 c8
A m2
A > c+8 c8 < a+8 g+8 g+4 g4 f4.

```


## バイナリデータ

本音楽ドライバ(sound.asm)が解釈可能なバイナリデータの仕様を以下に説明します。

### 構成

バイナリデータは、次の３種類のデータで構成されます。

* 音長テーブル
* シーケンスデータ

音長テーブルは曲全体で１つずつ存在します。<br>
シーケンスデータは、使用するチャンネル分存在します。

### 音長テーブル

音長テーブルは、演奏する各音の演奏時間を収集して
テーブル化したものです。

音長データは、キーオンとキーオフの各tickを8bit値のペアで定義します。<br>
四分音符のtickは48ですが、クォンタイズが7とした場合、
キーオンは42tick、キーオフは6tickになるため、
音長テーブルには次のように定義されます。

```
	.db	42, 6
```


### シーケンスデータ

シーケンスデータは、各チャンネルの発音を制御するコマンドが並んだデータ列です。<br>
本音楽ドライバは、このシーケンスデータを順に読み込み解釈することで
曲の演奏を行います。

シーケンスデータのコマンドは次の４種類に大きく別れます。
* 音符コマンド
* 休符コマンド
* 延長コマンド
* 特殊コマンド

上記コマンドの定義域は次の通りです。

```
     +0 +1 +2 +3 +4 +5 +6 +7 +8 +9 +A +B +C +D +E +F
    +-----------------------------------------------+
$00 |                  特殊コマンド                 |
    +--+--+-----------------------------------------+
$10 |  |  |                                         |
$20 |  |  |                                         |
$30 |  |  |                                         |
$40 |  |  |                                         |
$50 |  |  |                                         |
$60 |延|休|                                         |
$70 |長|符|                                         |
$80 |コ|コ|            音符コマンド                 |
$90 |マ|マ|                                         |
$A0 |ン|ン|                                         |
$B0 |ド|ド|                                         |
$C0 |  |  |                                         |
$D0 |  |  |                                         |
$E0 |  |  |                                         |
$F0 |  |  |                                         |
    +--+--+-----------------------------------------+
```

#### 音符コマンド

発音を指示するコマンドです。

コマンドは8bit長でフォーマットは次の通りです。

```
xxxxyyyy

x : 音長差分値
y : 音程差分値
```

コマンドの上位4bitの音長差分値(1～15)から1を引いて
音長ベース値を加算した値をインデックスとして、
音長テーブルからキーオン/キーオフのtickを表引きします。<br>
表引きしたtickの分の時間、発音します。

コマンドの下位4bitの音程差分値(2～15)から2を引いて
音程ベース値を加算した値を音程として発音します。<br>
音程は0～71の値でo2c～o7bの音程に対応します。

矩形波、および三角波では、音程をインデックスとして
音楽ドライバ内の周波数テーブルを表引きして周波数を求め、
レジスタに書き込みます。

ノイズは、音程が31以下の場合にはノイズ周波数指定、
32以上の場合にはリズム指定となります。

ノイズ周波数指定は、音程($00～$1f)から$10を引いて
$8fとANDをとった値($80～$8f,$00～$0f)をレジスタに書き込みます。<br>
したがって、音程(0～15 o2c～o3d+)が短周期ノイズ、
音程(16～31 o3e～o4g)が長周期ノイズとなります。

リズム指定は、音程($20～)から$20を引いた値をインデックスとして
音長テーブルに定義されたリズム定義を表引きし、
リズム定義の情報に従って、ノイズの音程や音量を設定します。

|リズム名|音程|
|---|---|
|無音|32(o4g+)|
|バスドラム|33(o4a)|
|ハイハット|34(o4a+)|
|ハイハット＋バスドラム|35(o4b)|

#### 休符コマンド

無音を指示するコマンドです。

コマンドは8bit長でフォーマットは次の通りです。

```
xxxx0001

x : 音長差分値
```

コマンドの上位4bitの音長差分値(1～15)から1を引いて
音長ベース値を加算した値をインデックスとして、
音長テーブルからキーオン/キーオフのtickを表引きします。<br>
表引きしたtickの分の時間、音を止めます。

#### 延長コマンド

発音または無音の時間を延長するコマンドです。

コマンドは8bit長でフォーマットは次の通りです。

```
xxxx0000

x : 音長差分値
```

コマンドの上位4bitの音長差分値(1～15)から1を引いて
音長ベース値を加算した値をインデックスとして、
音長テーブルからキーオン/キーオフのtickを表引きします。<br>
表引きしたtickをカウンタに加算し、発音または無音中の時間を延長します。

#### 特殊コマンド

発音や無音以外のシーケンス制御に必要なコマンドをまとめたグループです。

特殊コマンドの詳細を下記に示します。

|コマンド名|フォーマット|説明|
|---|---|---|
| 終端コマンド | $00 | シーケンスの終端を示すコマンドです。<br>チャンネルの演奏を停止します。|
| リターンコマンド | $01 | サブルーチン呼び出し中の場合は呼び出し元に戻ります。<br> サブルーチン呼び出し中でない場合は何も処理しません。 |
| サブルーチンコマンド | $02 $nn [ $nn ] | サブルーチン制御を行います。<br>現在のシーケンス位置を保存しジャンプします。<br>2バイトサイズのショート形式と3バイトサイズのロング形式の２種類があります。<br><br>ショート形式 (-64～63)<br>2バイト目7bit : 0<br>2バイト目6～0bit : ジャンプ先相対位置<br><br>ロング形式 (-16384～16383)<br>2バイト目7bit : 1<br>2バイト目6～0 bit : ジャンプ先相対位置 下位7bit<br>3バイト目7～0 bit : ジャンプ先相対位置 上位8bit |
| ジャンプコマンド | $03 $nn [ $nn ] | ジャンプ制御を行います。<br>2バイトサイズのショート形式と3バイトサイズのロング形式の２種類があります。<br>フォーマットはサブルーチンコマンドと同じです。 |
| リピートコマンド<br>(ショート形式) | $04 $nn | リピート制御を行います。<br><br>2バイト目 7～5bit : 繰り返し回数(2～9回)<br>2バイト目 4～0bit : 戻り位置(-32～-1バイト)|
| リピートコマンド<br>(ロング形式) | $05 $nn $nn | リピート制御を行います。<br><br>2バイト目 7～0bit : 繰り返し回数(2～127回)<br>3バイト目 7～0bit : 戻り位置(-256～-1バイト)|
| 広域リピートコマンド<br>(ショート形式) | $06 $nn | リピート制御を行います。<br><br>2バイト目 7～5bit : 繰り返し回数(2～9回)<br>2バイト目 4～0bit : 戻り位置(-32～-1バイト)|
| 広域リピートコマンド<br>(ロング形式) | $07 $nn $nn | リピート制御を行います。<br><br>2バイト目 7～0bit : 繰り返し回数(2～127回)<br>3バイト目 7～0bit : 戻り位置(-256～-1バイト)|
| レジスタ1設定コマンド | $08 $nn | $4000(矩形波1)、$4004(矩形波2)、$400c(ノイズ)のレジスタに値($nn)を設定します。 |
| レジスタ2設定コマンド | $09 $nn | $4001(矩形波1)、$4005(矩形波2)、$4009(三角波)、$400d(ノイズ)のレジスタに値($nn)を設定します。<br>値の保存領域を広域リピート機能のカウンタと共有しているため、カウンタの初期化としても使用できます。 |
| 音程ベース設定コマンド | $0a $nn | 音程ベース値($nn)を設定するコマンドです。 |
| 音長ベース設定コマンド | $0b $nn | 音長ベース値($nn)を設定するコマンドです。 |
| エンベロープ設定コマンド | $0c $nn | エンベロープデータを設定するコマンドです。 |
| テンポ設定コマンド | $0e $nn | テンポ($nn)を設定するコマンドです。 |
| ループカウンタコマンド | $0f | 曲のループ回数のカウンタを1増加します。|

